/*
 * bg_tree_path
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <bg.h>

int bg_tree_source_create(bg_tree_path_t *tp, bg_tree_t *parent,
    const ffchar_t *path)
{
  size_t len1=parent&&parent->source.path?FFSTRLEN(parent->source.path):0u;
  size_t len2=path?FFSTRLEN(path):0u;
  ffchar_t *pp;

  if (parent&&path) {
    // don't copy trailing path separators. ///////////////////////////////////
    while (0u<len2&&FFPATHSEP==path[len2-1])
      --len2;

    tp->path=pp=malloc(((len1?len1+1u:0u)+(len2?len2+1u:0u))*sizeof(pp[0]));

    if (!tp->path) {
      DMESSAGE("allocating path");
      goto epath;
    }

    if (parent->source.path) {
      // we're more than one plane below root and hence dirname is
      // provided by our parent and basename is path.
      memcpy(pp,parent->source.path,len1*sizeof *pp);
      pp+=len1;
      *pp++=FFPATHSEP;
      tp->basename=pp;
      memcpy(pp,path,len2*sizeof *pp);
      pp+=len2;
      *pp++=FFL('\0');
    }
    else {
      // we're at the first plane below root and hence need to extract
      // dirname and basename from path.
      memcpy(pp,path,len2*sizeof *pp);
      pp+=len2;
      *pp++=FFL('\0');
    }

    ///////////////////////////////////////////////////////////////////////////
    tp->basename=bg_basename(tp->path);
  }
  else {
    tp->path=NULL;
    tp->basename=NULL;
  }

  /////////////////////////////////////////////////////////////////////////////
  return 0;
//cleanup:
  if (tp->path)
    free(tp->path);
epath:
  return -1;
}

void bg_tree_path_destroy(bg_tree_path_t *tp)
{
  if (tp->path)
    free(tp->path);
}
