/*
 * bgx.c
 * Copyright (C) 2014-2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
/*
 * once meant as an example program on how to deal with libbg.a (and in
 * turn with libff.a) finally serving as bs1770gain's implementation.
 */
#if defined (_WIN32) // [
#include <fcntl.h>
#include <getoptW.h>
#else // ] [
#include <ctype.h>
#include <getopt.h>
#endif // ]
#include <locale.h>
#include <bg.h>

///////////////////////////////////////////////////////////////////////////////
#if 1 // [
#if 1 // [
#define BS1770GAIN_TEAM_WHITE \
    FFL(" #TeamWhite")
#else // ] [
#define BS1770GAIN_TEAM_WHITE \
    FFL("")
#endif // ]
#define BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES \
    FFL("Nanos gigantum humeris insidentes.") BS1770GAIN_TEAM_WHITE FFL("\n")
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (BG_VERSION) // [
static const ffchar_t *bg_version(const ffchar_t *path, FILE *f)
{
  path=FFBASENAME(path);

  FFFPUTS(path,f);
#if defined (PACKAGE_VERSION) // [
  FFFPUTC(FFL(' '),f);
  FFFPUTS(FFSTR(PACKAGE_VERSION),f);
#endif // ]
  FFFPUTS(FFL(", Copyright (C) Peter Belkner 2014-2019.\n"),f),
#if defined (BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES) // [
  FFFPUTS(BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES,f);
#endif // ]
#if defined (PACKAGE_URL) // [
  FFFPUTS(FFSTR(PACKAGE_URL),f);
  FFFPUTC(FFL('\n'),f);
#endif // ]

  return path;
}
#endif // ]

static void bg_usage(const ffchar_t *path, FILE *f)
{
#if defined (BG_VERSION) // [
  path=bg_version(path,f);
#else // ] [
  path=FFBASENAME(path);
  FFFPUTS(path,f);
#if defined (PACKAGE_VERSION) // [
  FFFPUTC(FFL(' '),f);
  FFFPUTS(FFSTR(PACKAGE_VERSION),f);
#endif // ]
  FFFPUTS(FFL(", Copyright (C) Peter Belkner 2014-2019.\n"),f),
#if defined (BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES) // [
  FFFPUTS(BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES,f);
#endif // ]
#if defined (PACKAGE_URL) // [
  FFFPUTS(FFSTR(PACKAGE_URL),f);
  FFFPUTC(FFL('\n'),f);
#endif // ]
#endif // ]
  FFFPUTC(FFL('\n'),f);
  FFFPRINTF(f,FFL("Usage:  %s [options] <file/dir> [<file/dir> ...]\n"),path);
  FFFPUTC(FFL('\n'),f);
  FFFPUTS(FFL("Options:\n"),f);
#if defined (BG_VERSION) // [
  FFFPUTS(FFL(" -v,--version:  print this message and exit\n"),f);
#endif // ]
  FFFPUTS(FFL(" -h,--help[=<topic>]?:  print this message and exit\n")
      FFL("    (optional argument <topic> out of \"codec\" or \"suffix\")\n"),f);
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTS(FFL(" -i,--integrated:  calculate integrated loudness\n"),f);
  FFFPUTS(FFL(" -s,--shortterm:  calculate maximum shortterm loudness\n"),f);
  FFFPUTS(FFL(" -m,--momentary:  calculate maximum momentary loudness\n"),f);
  FFFPUTS(FFL(" -r,--range:  calculate loudness range\n"),f);
  FFFPUTS(FFL(" -p,--samplepeak:  calculate maximum sample peak\n"),f);
  FFFPUTS(FFL(" -t,--truepeak:  calculate maximum true peak\n"),f);
  FFFPUTS(FFL(" -b <timestamp>,--begin=<timestamp>:  begin decoding at\n")
      FFL("   timestamp (in microseconds, format: hh:mm:ss.ms)\n"),f);
  FFFPUTS(FFL(" -d <duration>,--duration=<duration>:  let decoding\n")
      FFL("   last duration (in microseconds, format: hh:mm:ss.ms)\n"),f);
  FFFPUTS(FFL(" -u <method>,--use=<method>:  base replaygain calculation")
          FFL(" on\n")
      FFL("   <method> (with respect to the -a/--apply and -o/--output\n")
      FFL("   options. available methods:\n")
      FFL("     integrated (default),\n")
      FFL("     momentary, or\n")
      FFL("     shortterm\n")
      FFL("   experimental methods:\n")
      FFL("     momentary-mean (synonym for \"integrated\"),\n")
      FFL("     momentary-maximum (synonym for \"momentary\"),\n")
      FFL("     shortterm-mean, or\n")
      FFL("     shortterm-maximum (synonym for \"shortterm\")\n"),f);
  FFFPUTS(FFL(" -a [<weight>]/--apply[=<weight>]:  apply the EBU/ATSC/RG\n")
      FFL("   album gain (in conjunction with the -o/--output option.)\n")
      FFL("   when <weight> out of [0.0 .. 1.0] is provided: album gain\n")
      FFL("   plus <weight>*(track gain - album gain).\n"),f);
  FFFPUTS(FFL(" -o <folder>,--output=<folder>:  write replaygain tags\n")
      FFL("   or apply the EBU/ATSC/RG gain, respectively,\n")
      FFL("   and output to <folder>.\n")
      FFL("   specify either option -o/--output or option --overwrite\n")
      FFL("   but not both.\n"),f);
#if defined (_WIN32) // [
  FFFPUTS(FFL(" -f <file>,--file=<file>:  write analysis to an utf-8 log")
      FFL(" file\n"),f);
  FFFPUTS(FFL(" --utf-16[=<file>]:  write an utf-16 log file instead of")
      FFL (" an utf-8\n")
      FFL("   log file (<file> can be omitted when provided by means of the\n")
      FFL("   -f/--file option.)\n"),f);
#else // ] [
  FFFPUTS(FFL(" -f <file>,--file=<file>:  write analysis to a log file\n"),f);
#endif // ]
#if defined (BG_LIST) // [
  FFFPUTS(FFL(" -l,--list:  print FFmpeg format/stream information\n"),f);
#endif // ]
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTS(FFL(" --ebu:  calculate replay gain according to EBU R128\n")
      FFL("   (-23.0 LUFS, default)\n"),f);
  FFFPUTS(FFL(" --atsc:  calculate replay gain according to ATSC A/85\n")
      FFL("   (-24.0 LUFS)\n"),f);
  FFFPUTS(FFL(" --replaygain:  calculate replay gain according to\n")
      FFL("   ReplayGain 2.0 (-18.0 LUFS)\n"),f);
  FFFPUTS(FFL(" --track-tags:  write track tags\n"),f);
  FFFPUTS(FFL(" --album-tags:  write album tags\n"),f);
  FFFPUTS(FFL(" --tag-prefix=<prefix>:  instead of \"REPLAYGAIN\",\n")
      FFL("   use <prefix> as replaygain tag prefix\n"),f);
  FFFPUTS(FFL(" --unit=<unit>:  write results and tags with <unit> out of\n")
      FFL("   \"ebu\" or \"db\"\n"),f);
#if defined (BG_LIST) // [
  FFFPUTS(FFL(" --audio=<index>/--ai=<index>:  select audio index."),f);
#else // ] [
  FFFPUTS(FFL(" --audio=<index>/--ai=<index>:  select audio index")
      FFL(" (corresponds\n")
      FFL("   to [0:<index>] in FFmpeg listing, cf. -l/--list option)\n"),f);
#endif // ]
  FFFPUTS(FFL(" --video=<index>/--vi=<index>:  select video index")
      FFL(" (corresponds\n")
      FFL("   to [0:<index>] in FFmpeg listing, cf. -l/--list option)\n"),f);
#if defined (BG_PARAM_MATRIX) // [
  FFFPUTS(FFL(" --matrix:<matrix>:  remix to <matrix> out of\n")
      FFL("    front-left,\n")
      FFL("    front-right,\n")
      FFL("    front-center,\n")
      FFL("    low-frequency,\n")
      FFL("    back-left,\n")
      FFL("    back-right,\n")
      FFL("    front-left-of-center,\n")
      FFL("    front-right-of-center,\n")
      FFL("    back-center,\n")
      FFL("    side-left,\n")
      FFL("    side-right,\n")
      FFL("    top-center,\n")
      FFL("    top-front-left,\n")
      FFL("    top-front-center,\n")
      FFL("    top-front-right,\n")
      FFL("    top-back-left,\n")
      FFL("    top-back-center,\n")
      FFL("    top-back-right,\n")
      FFL("    stereo-left,\n")
      FFL("    stereo-right,\n")
      FFL("    wide-left,\n")
      FFL("    wide-right,\n")
      FFL("    surround-direct-left,\n")
      FFL("    surround-direct-right,\n")
      FFL("    low-frequency-2,\n")
      FFL("    layout-mono,\n")
      FFL("    layout-stereo,\n")
      FFL("    layout-2point1,\n")
      FFL("    layout-2-1,\n")
      FFL("    layout-surround,\n")
      FFL("    layout-3point1,\n")
      FFL("    layout-4point0,\n")
      FFL("    layout-4point1,\n")
      FFL("    layout-2-2,\n")
      FFL("    layout-quad,\n")
      FFL("    layout-5point0,\n")
      FFL("    layout-5point1,\n")
      FFL("    layout-5point0-back,\n")
      FFL("    layout-5point1-back,\n")
      FFL("    layout-6point0,\n")
      FFL("    layout-6point0-front,\n")
      FFL("    layout-hexagonal,\n")
      FFL("    layout-6point1,\n")
      FFL("    layout-6point1-back,\n")
      FFL("    layout-6point1-front,\n")
      FFL("    layout-7point0,\n")
      FFL("    layout-7point0-front,\n")
      FFL("    layout-7point1,\n")
      FFL("    layout-7point1-wide,\n")
      FFL("    layout-7point1-wide-back,\n")
      FFL("    layout-octagonal,\n")
      FFL("    layout-hexadecagonal, or\n")
      FFL("    layout-stereo-downmix\n"),f);
  FFFPUTS(FFL(" --stereo:  shorthand for --matrix=layout-stereo\n"),f);
#else // ] [
  FFFPUTS(FFL(" --stereo:  convert to stereo\n"),f);
#endif // ]
  FFFPUTS(FFL(" --drc=<float>:  set AC3 dynamic range compression (DRC)\n"),f);
  FFFPUTS(FFL(" -x [<extension>]?, --extension[=<extension>]?:  enable")
      FFL(" extension\n")
      FFL("   out of\n")
      FFL("    rename:  rename files according to TITLE tag\n")
      FFL("    csv:  read metadata from per-folder CSV file")
          FFL(" \"folder.csv\"\n")
      FFL("    copy:  copy non-audio files from source to destination\n")
      FFL("      folder\n")
      FFL("    tags:  automatically add the TRACK and DISC tags\n")
      FFL("    all:  all of the above (can be omitted)\n"),f);
  FFFPUTS(FFL(" --suffix=<suffix>:  output to <basename>.<suffix>\n")
      FFL("    (only useful in conjunction with option -o/--output))\n"),f);
  FFFPUTS(FFL(" --loglevel=<level>:  set FFmpeg loglevel to <level> out of\n")
      FFL("    quiet,\n")
      FFL("    panic,\n")
      FFL("    fatal,\n")
      FFL("    error,\n")
      FFL("    warning,\n")
      FFL("    info,\n")
      FFL("    verbose,\n")
      FFL("    debug, or\n")
      FFL("    trace\n"),f);
  FFFPUTS(FFL(" --xml:  print results in xml format\n"),f);
  FFFPUTS(FFL(" --time:  print out duration of program invocation\n"),f);
  FFFPUTS(FFL(" --norm=<float>:  norm loudness to float.\n"),f);
#if 0 // [
  FFFPUTS(FFL(" --preamp=<preamp>:\n"),f);
  FFFPUTS(FFL(" --stero:\n"),f);
  FFFPUTS(FFL(" --rg-tags:\n"),f);
  FFFPUTS(FFL(" --bwf-tags:\n"),f);
#endif // ]
  FFFPUTS(FFL(" --overwrite:  replace any source file by its respective\n")
      FFL("    target file. specify either option -o/--output or option\n")
      FFL("    --overwrite but not both.\n")
      FFL("    WARNING:  use this option with extrem care. The source\n")
      FFL("      files will definitely be lost!!! For this reason using\n")
      FFL("      this option is discouraged. It's provided just for\n")
      FFL("      completeness.\n"),f);
#if defined (BG_CHANNEL_LFE) // [
  FFFPRINTF(f,FFL(" --lfe=<lfe channel>  overwrite the default <lfe")
      FFL(" channel>\n")
      FFL("    (default: %d.)\n"),BG_CHANNEL_LFE);
#endif // ]
  FFFPUTS(FFL(" --codec=<name>:  use audio codec \"<name>\" for output\n")
      FFL("    default: \"flac\".)\n"),f);
  FFFPUTS(FFL(" --temp-prefix=<prefix>:  create temporary files with\n")
      FFL("    prefix \"<prefix>\" default: \"")
      BG_TEMP_PREFIX
      FFL("\".)\n"),f);
  FFFPUTS(FFL(" --suppress-hierarchy:  suppress printing results up the\n")
      FFL("    hierarchy.\n"),f);
  FFFPUTS(FFL(" --suppress-progess:  suppress printing processing")
      FFL(" progress\n"),f);
#if defined (BG_PARAM_SEQUENTIAL) // [
  FFFPUTS(FFL(" --sequential:  fallback to the memory exhausting classical\n")
      FFL("    processing strategy.\n"),f);
#endif // ]
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTC(FFL('\n'),f);
  FFFPUTS(FFL("Experimental options:\n"),f);
  ////////
  FFFPUTS(FFL("1) momentary block\n"),f);
  FFFPUTS(FFL(" --momentary-mean:  calculate mean loudness based on\n")
      FFL("   momentary block (same as --integrated)\n"),f);
  FFFPUTS(FFL(" --momentary-maximum:  calculate maximum loudness based\n")
      FFL("   on momentary block (same as --momentary)\n"),f);
  FFFPUTS(FFL(" --momentary-range:  calculate loudness range based on\n")
      FFL("   momentary block\n"),f);
  FFFPUTS(FFL(" --momentary-length=<ms>:  length of momentary block\n")
      FFL("   in milliseconds (default: 400)\n"),f);
  FFFPUTS(FFL(" --momentary-overlap=<percent>:  overlap of momentary\n")
      FFL("   block in percent (default: 75)\n"),f);
  FFFPUTS(FFL(" --momentary-mean-gate=<gate>:  silence gate for mean\n")
      FFL("   measurement of momentary block (default: -10.0)\n"),f);
  FFFPUTS(FFL(" --momentary-range-gate=<gate>:  silence gate for range\n")
      FFL("   measurement of momentary block (default: -20.0)\n"),f);
  FFFPUTS(FFL(" --momentary-range-lower-bound=<float>:  lower bound for\n")
      FFL("   range measurement of momentary block (default: 0.1)\n"),f);
  FFFPUTS(FFL(" --momentary-range-upper-bound=<float>:  upper bound for\n")
      FFL("   range measurement of momentary block (default: 0.95)\n"),f);
  ////////
  FFFPUTS(FFL("2) shortterm block\n"),f);
  FFFPUTS(FFL(" --shortterm-mean:  calculate mean loudness based on\n")
      FFL("   shortterm block\n"),f);
  FFFPUTS(FFL(" --shortterm-maximum:  calculate maximum loudness based\n")
      FFL("   on shortterm block (same as --shortterm)\n"),f);
  FFFPUTS(FFL(" --shortterm-range:  calculate loudness range based on\n")
      FFL("   shortterm block (same as --range)\n"),f);
  FFFPUTS(FFL(" --shortterm-length=<ms>:  length of shortterm block\n")
      FFL("   in milliseconds (default: 3000)\n"),f);
  FFFPUTS(FFL(" --shortterm-overlap <percent>:  overlap of shortterm\n")
      FFL("   block in percent (default: 67)\n"),f);
  FFFPUTS(FFL(" --shortterm-mean-gate=<gate>:  silence gate for mean\n")
      FFL("   measurement of shortterm block (default: -10.0)\n"),f);
  FFFPUTS(FFL(" --shortterm-range-gate=<gate>:  silence gate for range\n")
      FFL("   measurement of shortterm block (default: -20.0)\n"),f);
  FFFPUTS(FFL(" --shortterm-range-lower-bound=<float>:  lower bound for\n")
      FFL("   range measurement of shortterm block (default: 0.1)\n"),f);
  FFFPUTS(FFL(" --shortterm-range-upper-bound=<float>:  upper bound for\n")
      FFL("   range measurement of shortterm block (default: 0.95)\n"),f);
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTS(FFL("\n"),f);
  FFFPUTS(FFL("Command line arguments may appear in any order.\n"),f);
}

///////////////////////////////////////////////////////////////////////////////
#define BG_PARAM_STRUCT_INITIALIZER
// https://stackoverflow.com/questions/190543/how-can-i-change-the-width-of-a-windows-console-window
// https://docs.microsoft.com/en-us/windows/console/window-and-screen-buffer-size
// David Tran: Logging Function Entry and Exit in C
//  https://davidtranscend.com/blog/log-function-entry-exit-c/
#if defined (_WIN32) // [
int wmain(int argc, wchar_t *const *argv)
#else // ] [
int main(int argc, char *const *argv)
#endif // ]
{
  enum {
#if 0 // [
    BG_ARG_APPLY=FFL('z')+1,
    BG_ARG_UNIT,
#else // ] [
    BG_ARG_UNIT=FFL('z')+1,
#endif // ]
#if defined (_WIN32) // [
    BG_ARG_UTF_16,
#endif // ]
    BG_ARG_DRC,
    BG_ARG_LOGLEVEL,
    BG_ARG_NORM,
    BG_ARG_PREAMP,
    BG_ARG_TIME,
    BG_ARG_EBU,
    BG_ARG_SUFFIX,
    BG_ARG_ATSC,
    BG_ARG_AUDIO,
    BG_ARG_VIDEO,
#if defined (BG_PARAM_MATRIX) // [
    BG_ARG_MATRIX,
#endif // ]
    BG_ARG_STEREO,
    BG_ARG_REPLAYGAIN,
    BG_ARG_RG_TAGS,
#if defined (BG_BWF_TAGS) // [
    BG_ARG_BWF_TAGS,
#endif // ]
    BG_ARG_TAGS_TRACK,
    BG_ARG_TAGS_ALBUM,
    BG_ARG_XML,
    BG_ARG_TAG_PREFIX,
    BG_ARG_OVERWRITE,
    BG_ARG_CH_LFE,
    BG_ARG_OUTPUT_CODEC,
    BG_ARG_TEMP_PREFIX,
    BG_ARG_SUPPRESS_HIERARCHY,
    BG_ARG_SUPPRESS_PROGRESS,
    BG_ARG_EXTENSION,
#if defined (BG_PARAM_SEQUENTIAL) // [
    BG_ARG_SEQUENTIAL,
#endif // ]
    ////////////////
#if 0 // [
    BG_ARG_MOMENTARY_MEAN,
    BG_ARG_MOMENTARY_MAXIMUM,
#endif // ]
    BG_ARG_MOMENTARY_RANGE,
    BG_ARG_MOMENTARY_LENGTH,
    BG_ARG_MOMENTARY_OVERLAP,
    BG_ARG_MOMENTARY_MEAN_GATE,
    BG_ARG_MOMENTARY_RANGE_GATE,
    BG_ARG_MOMENTARY_RANGE_LOWER_BOUND,
    BG_ARG_MOMENTARY_RANGE_UPPER_BOUND,
    ////////////////
    BG_ARG_SHORTTERM_MEAN,
#if 0 // [
    BG_ARG_SHORTTERM_MAXIMUM,
    BG_ARG_SHORTTERM_RANGE,
#endif // ]
    BG_ARG_SHORTTERM_LENGTH,
    BG_ARG_SHORTTERM_OVERLAP,
    BG_ARG_SHORTTERM_MEAN_GATE,
    BG_ARG_SHORTTERM_RANGE_GATE,
    BG_ARG_SHORTTERM_RANGE_LOWER_BOUND,
    BG_ARG_SHORTTERM_RANGE_UPPER_BOUND,
  };

  enum BGFlagHelpArg {
    BG_FLAG_HELP_ARG_NULL=1<<0,
    BG_FLAG_HELP_ARG_SUFFIX=1<<1,
    BG_FLAG_HELP_ARG_CODEC=1<<2,
  };

#if defined (HAVE_FF_DYNLOAD) // [
  static const ffchar_t TOOLS[]=FFL("bs1770gain-tools");
#endif // ]
  static const ffchar_t FLAGS[]=
#if defined (BG_VERSION) // [
      FFL("v")
#endif // ]
      FFL("u:ptismr")
#if 0 && defined (_WIN32) // [
      FFL("v")
#endif // ]
      FFL("b:d:o:f:ahlx");

  static struct option OPTS[]={
#if defined (BG_VERSION) // [
    { FFL("version"),no_argument,NULL,
        FFL('v') },
#endif // ]
    { FFL("help"),optional_argument,NULL,
        FFL('h') },
#if defined (BG_LIST) // [
    { FFL("list"),no_argument,NULL,
        FFL('l') },
#endif // ]
    { FFL("begin"),required_argument,NULL,
        FFL('b') },
    { FFL("duration"),required_argument,NULL,
        FFL('d') },
    { FFL("file"),required_argument,NULL,
        FFL('f') },
    { FFL("utf-8"),required_argument,NULL,
        FFL('f') },
    { FFL("output"),required_argument,NULL,
        FFL('o') },
    { FFL("use"),required_argument,NULL,
        FFL('u') },
    { FFL("integrated"),no_argument,NULL,
        FFL('i') },
    { FFL("momentary"),no_argument,NULL,
        FFL('m') },
    { FFL("range"),no_argument,NULL,
        FFL('r') },
    { FFL("samplepeak"),no_argument,NULL,
        FFL('p') },
    { FFL("shortterm"),no_argument,NULL,
        FFL('s') },
    { FFL("truepeak"),no_argument,NULL,
        FFL('t') },
    { FFL("apply"),optional_argument,NULL,
        FFL('a') },
#if defined (_WIN32) // [
    { FFL("utf-16"),optional_argument,NULL,
        BG_ARG_UTF_16 },
#endif // ]
    { FFL("unit"),required_argument,NULL,
        BG_ARG_UNIT },
    { FFL("audio"),required_argument,NULL,
        BG_ARG_AUDIO },
    { FFL("ai"),required_argument,NULL,
        BG_ARG_AUDIO },
    { FFL("drc"),required_argument,NULL,
        BG_ARG_DRC },
#if 0 // [
    { FFL("extension"),no_argument,NULL,
        FFL('x') },
#else // ] [
    { FFL("extension"),optional_argument,NULL,
        BG_ARG_EXTENSION },
#endif // ]
#if defined (BG_PARAM_SEQUENTIAL) // [
    { FFL("sequential"),no_argument,NULL,
        BG_ARG_SEQUENTIAL },
#endif // ]
    { FFL("suffix"),required_argument,NULL,
        BG_ARG_SUFFIX },
    { FFL("loglevel"),required_argument,NULL,
        BG_ARG_LOGLEVEL },
    { FFL("norm"),required_argument,NULL,
        BG_ARG_NORM },
    { FFL("preamp"),required_argument,NULL,
        BG_ARG_PREAMP },
    { FFL("video"),required_argument,NULL,
        BG_ARG_VIDEO },
    { FFL("vi"),required_argument,NULL,
        BG_ARG_VIDEO },
    { FFL("time"),no_argument,NULL,
        BG_ARG_TIME },
    { FFL("ebu"),no_argument,NULL,
        BG_ARG_EBU },
    { FFL("atsc"),no_argument,NULL,
        BG_ARG_ATSC },
#if defined (BG_PARAM_MATRIX) // [
    { FFL("matrix"),required_argument,NULL,
        BG_ARG_MATRIX },
#endif // ]
    { FFL("stereo"),no_argument,NULL,
        BG_ARG_STEREO },
    { FFL("replaygain"),no_argument,NULL,
        BG_ARG_REPLAYGAIN },
#if defined (BG_BWF_TAGS) // [
    { FFL("rg-tags"),no_argument,NULL,
        BG_ARG_TAGS_RG },
    { FFL("bwf-tags"),no_argument,NULL,
        BG_ARG_TAGS_BWF },
#endif // ]
    { FFL("track-tags"),no_argument,NULL,
        BG_ARG_TAGS_TRACK },
    { FFL("album-tags"),no_argument,NULL,
        BG_ARG_TAGS_ALBUM },
    { FFL("xml"),no_argument,NULL,
        BG_ARG_XML },
    { FFL("tag-prefix"),required_argument,NULL,
        BG_ARG_TAG_PREFIX },
    { FFL("overwrite"),no_argument,NULL,
        BG_ARG_OVERWRITE },
    { FFL("lfe"),required_argument,NULL,
        BG_ARG_CH_LFE },
    { FFL("codec"),required_argument,NULL,
        BG_ARG_OUTPUT_CODEC },
    { FFL("temp-prefix"),required_argument,NULL,
        BG_ARG_TEMP_PREFIX },
    { FFL("suppress-hierarchy"),no_argument,NULL,
        BG_ARG_SUPPRESS_HIERARCHY },
    { FFL("suppress-progress"),no_argument,NULL,
        BG_ARG_SUPPRESS_PROGRESS },
    /// momentary /////////////////////////////////////////////////////////////
    { FFL("momentary-mean"),no_argument,NULL,
        FFL('i') },
    { FFL("momentary-maximum"),no_argument,NULL,
        FFL('m') },
    { FFL("momentary-range"),no_argument,NULL,
        BG_ARG_MOMENTARY_RANGE },
    { FFL("momentary-length"),required_argument,NULL,
        BG_ARG_MOMENTARY_LENGTH },
    { FFL("momentary-overlap"),required_argument,NULL,
        BG_ARG_MOMENTARY_OVERLAP },
    { FFL("momentary-mean-gate"),required_argument,NULL,
        BG_ARG_MOMENTARY_MEAN_GATE },
    { FFL("momentary-range-gate"),required_argument,NULL,
        BG_ARG_MOMENTARY_RANGE_GATE },
    { FFL("momentary-range-lower-bound"),required_argument,NULL,
        BG_ARG_MOMENTARY_RANGE_LOWER_BOUND },
    { FFL("momentary-range-upper-bound"),required_argument,NULL,
        BG_ARG_MOMENTARY_RANGE_UPPER_BOUND },
    { FFL("shortterm-mean"),no_argument,NULL,
        BG_ARG_SHORTTERM_MEAN },
    /// shortterm /////////////////////////////////////////////////////////////
    { FFL("shortterm-maximum"),no_argument,NULL,
        FFL('s') },
    { FFL("shortterm-range"),no_argument,NULL,
        FFL('r') },
    { FFL("shortterm-length"),required_argument,NULL,
        BG_ARG_SHORTTERM_LENGTH },
    { FFL("shortterm-overlap"),required_argument,NULL,
        BG_ARG_SHORTTERM_OVERLAP },
    { FFL("shortterm-mean-gate"),required_argument,NULL,
        BG_ARG_SHORTTERM_MEAN_GATE },
    { FFL("shortterm-range-gate"),required_argument,NULL,
        BG_ARG_SHORTTERM_RANGE_GATE },
    { FFL("shortterm-range-lower-bound"),required_argument,NULL,
        BG_ARG_SHORTTERM_RANGE_LOWER_BOUND },
    { FFL("shortterm-range-upper-bound"),required_argument,NULL,
        BG_ARG_SHORTTERM_RANGE_UPPER_BOUND },
    { 0 }
  };

  int code=1;
#if defined (_WIN32) // [
  int utf16=0;
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM.
  const char *lang=getenv("LANG");
#else // ] [
  const char *locale=NULL;
#endif // ]
  const ffchar_t *fpath=NULL;
  enum BGFlagHelpArg help_args=0;
  bg_visitor_t analyzer;
  bg_visitor_t muxer;
  bg_tree_t root;
  bg_param_t param;
  int c;
  double overlap;
  unsigned version;
#if 0 && defined (BG_PARAM_MATRIX) // [
  ffchar_t *op;
#endif // ]
#if (58<=LIBAVCODEC_VERSION_MAJOR||58<=LIBAVFORMAT_VERSION_MAJOR) // [
  void *opaque;
#endif // ]
  const AVCodec *codec;
  const AVOutputFormat *oformat;

  /////////////////////////////////////////////////////////////////////////////
  memset(&analyzer,0,sizeof analyzer);
  memset(&muxer,0,sizeof muxer);
  memset(&root,0,sizeof root);

  /////////////////////////////////////////////////////////////////////////////
  if (argc<2) {
    FFMESSAGE(FFL("missing command line arguments"));
    bg_usage(argv[0],stderr);
    goto e_arg;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bg_param_create(&param)<0) {
    DMESSAGE("create parameters");
    goto e_arg;
  }

  /////////////////////////////////////////////////////////////////////////////
  while (0<=(c=FF_GETOPT_LONG(argc,argv,FLAGS,OPTS,NULL))) {
    switch (c) {
    ///////////////////////////////////////////////////////////////////////////
    case FFL('?'):
      FFFPUTC(FFL('\n'),stderr);
      bg_usage(argv[0],stderr);
      goto e_arg;
    ///////////////////////////////////////////////////////////////////////////
#if defined (BG_VERSION) // [
    case FFL('v'):
#if 0 // [
        FFFPUTS(FFBASENAME(argv[0]),stdout);
#if defined (PACKAGE_VERSION) // [
        FFFPUTC(FFL(' '),stdout);
        FFFPUTS(FFSTR(PACKAGE_VERSION),stdout);
#endif // ]
#if defined (BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES) // [
        FFFPUTS(BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES,stdout);
#endif // ]
        FFFPUTC(FFL('\n'),stdout);
#else // ] [
        bg_version(argv[0],stdout);
#endif // ]
        code=0;
        goto e_arg;
#endif // ]
    case FFL('b'):
      param.interval.begin=bg_parse_time(optarg);
      break;
    case FFL('d'):
      param.interval.duration=bg_parse_time(optarg);
      break;
    case FFL('f'):
      fpath=optarg;
      break;
    case FFL('o'):
      if (!optarg) {
        DMESSAGE("missing argument to option -o/--output");
        goto e_arg;
      }
      else if (param.overwrite) {
        DMESSAGE("specify either option -o/--output or option --overwrite"
            " but not both");
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      param.output.dirname=bg_pathnorm(optarg);
      break;
    case FFL('u'):
      if (!FFSTRCASECMP(FFL("integrated"),optarg)
          ||!FFSTRCASECMP(FFL("momentary-mean"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MEAN;
      else if (!FFSTRCASECMP(FFL("momentary"),optarg)
          ||!FFSTRCASECMP(FFL("momentary-maximum"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MAXIMUM;
      else if (!FFSTRCASECMP(FFL("shortterm-mean"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MEAN;
      else if (!FFSTRCASECMP(FFL("shortterm"),optarg)
          ||!FFSTRCASECMP(FFL("shortterm-maximum"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MAXIMUM;
      else {
        FFVMESSAGE(FFL("method \"%s\" not recognized"),optarg);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    /// without argument ////////////////////////////////////////////////////
    case FFL('a'):
      param.flags.mode|=BG_FLAGS_MODE_APPLY;

      if (optarg) {
        param.weight.enabled=1;
        param.weight.value=FFATOF(optarg);

        if (param.weight.value<0.0||1.0<param.weight.value) {
          DMESSAGE("weight out of range");
          bg_usage(argv[0],stderr);
          goto e_arg;
        }
      }

      break;
    case FFL('h'):
      if (!optarg)
        help_args|=BG_FLAG_HELP_ARG_NULL;
      else if (!FFSTRCMP(FFL("codec"),optarg))
        help_args|=BG_FLAG_HELP_ARG_CODEC;
      else if (!FFSTRCMP(FFL("suffix"),optarg))
        help_args|=BG_FLAG_HELP_ARG_SUFFIX;
      else {
        FFFPRINTF(stderr,FFL("Error: help topic \"%s\" not recognized.\n\n"),
            optarg);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
#if defined (BG_LIST) // [
    case FFL('l'):
      param.result.f=stderr;
      param.dump=1;
      break;
#endif // ]
    case FFL('i'):
      param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MEAN;
      break;
    case FFL('s'):
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MAXIMUM;
      break;
    case FFL('m'):
      param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MAXIMUM;
      break;
    case FFL('r'):
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_RANGE;
      break;
    case FFL('p'):
      param.flags.aggregate|=BG_FLAGS_AGG_SAMPLEPEAK;
      break;
    case FFL('t'):
      param.flags.aggregate|=BG_FLAGS_AGG_TRUEPEAK;
      break;
    case FFL('x'):
    case BG_ARG_EXTENSION:
      if (!optarg) {
        param.flags.aggregate|=BG_FLAGS_AGG_TRUEPEAK;
        param.flags.extension=BG_FLAGS_EXT_ALL;
#if defined (BG_PARAM_STEREO) // [
        param.stereo=1;
#else // ] [
        param.decode.request.channel_layout=AV_CH_LAYOUT_STEREO;
#endif // ]
        param.decode.drc.enabled=1;
#if 0 // [
        param.decode.drc.scale=0.0;
#endif // ]
      }
      else if (0==FFSTRCASECMP(FFL("rename"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_RENAME;
      else if (0==FFSTRCASECMP(FFL("csv"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_CSV;
      else if (0==FFSTRCASECMP(FFL("copy"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_COPY;
      else if (0==FFSTRCASECMP(FFL("tags"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_TAGS;
      else if (0==FFSTRCASECMP(FFL("all"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_ALL;
      else {
        FFVMESSAGE(FFL("extension \"%s\" not recognized"),optarg);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    /// without flag ////////////////////////////////////////////////////////
    case BG_ARG_AUDIO:
      param.ai=FFATOI(optarg);
      break;
    case BG_ARG_VIDEO:
      param.vi=FFATOI(optarg);
      break;
    case BG_ARG_UNIT:
      if (!FFSTRCMP(optarg,FFL("ebu")))
        bg_param_set_unit_ebu(&param);
      else if (!FFSTRCMP(optarg,FFL("db")))
        bg_param_set_unit_db(&param);
      else {
        FFVMESSAGE(FFL("argument to option --unit not recognized: \"%s\""),
            optarg);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
#if defined (_WIN32) // [
    case BG_ARG_UTF_16:
      utf16=1;

      if (optarg)
        fpath=optarg;

      break;
#endif // ]
    case BG_ARG_DRC:
      param.decode.drc.enabled=1;
      param.decode.drc.scale=FFATOF(optarg);
      break;
    case BG_ARG_SUFFIX:
      param.out.sfx=optarg;
      break;
    case BG_ARG_TAG_PREFIX:
#if defined (_WIN32) // [
      ff_wcs2str(optarg,param.tag.pfx,CP_UTF8,(sizeof param.tag.pfx)-2);
#else // ] [
      strncpy(param.tag.pfx,optarg,(sizeof param.tag.pfx)-2);
#endif // ]
      param.tag.pfx[strlen(param.tag.pfx)+1]='\0';
      param.tag.pfx[strlen(param.tag.pfx)]='_';
      break;
    case BG_ARG_OVERWRITE:
      if (param.output.dirname) {
        DMESSAGE("specify either option -o/--output or option --overwrite"
            " but not both");
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      param.overwrite=1;
      break;
    case BG_ARG_CH_LFE:
      param.lfe=FFATOI(optarg);
      break;
    case BG_ARG_OUTPUT_CODEC:
#if defined (_WIN32) // [
      ff_wcs2str(optarg,param.codec.name,CP_UTF8,(sizeof param.codec.name)-1);
#else // ] [
      param.codec.name=optarg;
#endif // ]
      break;
    case BG_ARG_TEMP_PREFIX:
      if (!*optarg) {
        DMESSAGE("missing argument to --temp-prefix");
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      param.temp_prefix=optarg;
      break;
    case BG_ARG_SUPPRESS_HIERARCHY:
      param.suppress.hierarchy=1;
      break;
    case BG_ARG_SUPPRESS_PROGRESS:
      param.suppress.progress=1;
      break;
    case BG_ARG_LOGLEVEL:
      if (0==FFSTRCASECMP(FFL("quiet"),optarg))
        param.loglevel=AV_LOG_QUIET;
      else if (0==FFSTRCASECMP(FFL("panic"),optarg))
        param.loglevel=AV_LOG_PANIC;
      else if (0==FFSTRCASECMP(FFL("fatal"),optarg))
        param.loglevel=AV_LOG_FATAL;
      else if (0==FFSTRCASECMP(FFL("error"),optarg))
        param.loglevel=AV_LOG_ERROR;
      else if (0==FFSTRCASECMP(FFL("warning"),optarg))
        param.loglevel=AV_LOG_WARNING;
      else if (0==FFSTRCASECMP(FFL("info"),optarg))
        param.loglevel=AV_LOG_INFO;
      else if (0==FFSTRCASECMP(FFL("verbose"),optarg))
        param.loglevel=AV_LOG_VERBOSE;
      else if (0==FFSTRCASECMP(FFL("debug"),optarg))
        param.loglevel=AV_LOG_DEBUG;
      else if (0==FFSTRCASECMP(FFL("trace"),optarg))
        param.loglevel=AV_LOG_TRACE;
      else {
        FFFPRINTF(stderr,FFL("Error: loglevel \"%s\" not recognized.\n\n"),
            optarg);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    case BG_ARG_NORM:
      param.flags.norm=BG_FLAGS_NORM_NULL;
      param.norm=FFATOF(optarg);
      break;
    case BG_ARG_PREAMP:
      param.preamp=FFATOF(optarg);
      break;
    /// without flag ////////////////////////////////////////////////////////
    case BG_ARG_ATSC:
      param.flags.norm=BG_FLAGS_NORM_ATSC;
      param.norm=-24.0;
      bg_param_set_unit_db(&param);
      break;
    case BG_ARG_EBU:
      param.flags.norm=BG_FLAGS_NORM_EBU;
      param.norm=-23.0;
      bg_param_set_unit_ebu(&param);
      break;
    case BG_ARG_REPLAYGAIN:
      param.flags.norm=BG_FLAGS_NORM_REPLAYGAIN;
      param.norm=-18.0;
      bg_param_set_unit_db(&param);
      break;
#if defined (BG_BWF_TAGS) // [
    case BG_ARG_TAGS_RG:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_RG;
      break;
    case BG_ARG_TAGS_BWF:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_BWF;
      break;
#endif // ]
    case BG_ARG_TAGS_TRACK:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_TRACK;
      break;
    case BG_ARG_TAGS_ALBUM:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_ALBUM;
      break;
    case BG_ARG_XML:
      param.print.vtbl=&bg_print_xml_vmt;
      break;
    case BG_ARG_TIME:
      param.time=1;
      break;
#if defined (BG_PARAM_MATRIX) // [
    case BG_ARG_MATRIX:
#if 0 // [
      for (op=optarg;op&&*op;++op) {
        if (FFL('_')==*op)
          *op=FFL('-');
      }
#endif // ]
      if (!FFSTRCASECMP(FFL("front-left"),optarg))
        param.decode.request.channel_layout=AV_CH_FRONT_LEFT;
      else if (!FFSTRCASECMP(FFL("front-right"),optarg))
        param.decode.request.channel_layout=AV_CH_FRONT_RIGHT;
      else if (!FFSTRCASECMP(FFL("front-center"),optarg))
        param.decode.request.channel_layout=AV_CH_FRONT_CENTER;
      else if (!FFSTRCASECMP(FFL("low-frequency"),optarg))
        param.decode.request.channel_layout=AV_CH_LOW_FREQUENCY;
      else if (!FFSTRCASECMP(FFL("back-left"),optarg))
        param.decode.request.channel_layout=AV_CH_BACK_LEFT;
      else if (!FFSTRCASECMP(FFL("back-right"),optarg))
        param.decode.request.channel_layout=AV_CH_BACK_RIGHT;
      else if (!FFSTRCASECMP(FFL("front-left-of-center"),optarg))
        param.decode.request.channel_layout=AV_CH_FRONT_LEFT_OF_CENTER;
      else if (!FFSTRCASECMP(FFL("front-right-of-center"),optarg))
        param.decode.request.channel_layout=AV_CH_FRONT_RIGHT_OF_CENTER;
      else if (!FFSTRCASECMP(FFL("back-center"),optarg))
        param.decode.request.channel_layout=AV_CH_BACK_CENTER;
      else if (!FFSTRCASECMP(FFL("side-left"),optarg))
        param.decode.request.channel_layout=AV_CH_SIDE_LEFT;
      else if (!FFSTRCASECMP(FFL("side-right"),optarg))
        param.decode.request.channel_layout=AV_CH_SIDE_RIGHT;
      else if (!FFSTRCASECMP(FFL("top-center"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_CENTER;
      else if (!FFSTRCASECMP(FFL("top-front-left"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_FRONT_LEFT;
      else if (!FFSTRCASECMP(FFL("top-front-center"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_FRONT_CENTER;
      else if (!FFSTRCASECMP(FFL("top-front-right"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_FRONT_RIGHT;
      else if (!FFSTRCASECMP(FFL("top-back-left"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_BACK_LEFT;
      else if (!FFSTRCASECMP(FFL("top-back-center"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_BACK_CENTER;
      else if (!FFSTRCASECMP(FFL("top-back-right"),optarg))
        param.decode.request.channel_layout=AV_CH_TOP_BACK_RIGHT;
      else if (!FFSTRCASECMP(FFL("stereo-left"),optarg))
        param.decode.request.channel_layout=AV_CH_STEREO_LEFT;
      else if (!FFSTRCASECMP(FFL("stereo-right"),optarg))
        param.decode.request.channel_layout=AV_CH_STEREO_RIGHT;
      else if (!FFSTRCASECMP(FFL("wide-left"),optarg))
        param.decode.request.channel_layout=AV_CH_WIDE_LEFT;
      else if (!FFSTRCASECMP(FFL("wide-right"),optarg))
        param.decode.request.channel_layout=AV_CH_WIDE_RIGHT;
      else if (!FFSTRCASECMP(FFL("surround-direct-left"),optarg))
        param.decode.request.channel_layout=AV_CH_SURROUND_DIRECT_LEFT;
      else if (!FFSTRCASECMP(FFL("surround-direct-right"),optarg))
        param.decode.request.channel_layout=AV_CH_SURROUND_DIRECT_RIGHT;
      else if (!FFSTRCASECMP(FFL("low-frequency-2"),optarg))
        param.decode.request.channel_layout=AV_CH_LOW_FREQUENCY_2;
      else if (!FFSTRCASECMP(FFL("layout-mono"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_MONO;
      else if (!FFSTRCASECMP(FFL("layout-stereo"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_STEREO;
      else if (!FFSTRCASECMP(FFL("layout-2point1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_2POINT1;
      else if (!FFSTRCASECMP(FFL("layout-2-1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_2_1;
      else if (!FFSTRCASECMP(FFL("layout-surround"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_SURROUND;
      else if (!FFSTRCASECMP(FFL("layout-3point1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_3POINT1;
      else if (!FFSTRCASECMP(FFL("layout-4point0"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_4POINT0;
      else if (!FFSTRCASECMP(FFL("layout-4point1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_4POINT1;
      else if (!FFSTRCASECMP(FFL("layout-2-2"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_2_2;
      else if (!FFSTRCASECMP(FFL("layout-quad"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_QUAD;
      else if (!FFSTRCASECMP(FFL("layout-5point0"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_5POINT0;
      else if (!FFSTRCASECMP(FFL("layout-5point1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_5POINT1;
      else if (!FFSTRCASECMP(FFL("layout-5point0-back"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_5POINT0_BACK;
      else if (!FFSTRCASECMP(FFL("layout-5point1-back"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_5POINT1_BACK;
      else if (!FFSTRCASECMP(FFL("layout-6point0"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_6POINT0;
      else if (!FFSTRCASECMP(FFL("layout-6point0-front"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_6POINT0_FRONT;
      else if (!FFSTRCASECMP(FFL("layout-hexagonal"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_HEXAGONAL;
      else if (!FFSTRCASECMP(FFL("layout-6point1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_6POINT1;
      else if (!FFSTRCASECMP(FFL("layout-6point1-back"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_6POINT1_BACK;
      else if (!FFSTRCASECMP(FFL("layout_6point1_front"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_6POINT1_FRONT;
      else if (!FFSTRCASECMP(FFL("layout-7point0"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_7POINT0;
      else if (!FFSTRCASECMP(FFL("layout-7point0-front"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_7POINT0_FRONT;
      else if (!FFSTRCASECMP(FFL("layout-7point1"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_7POINT1;
      else if (!FFSTRCASECMP(FFL("layout-7point1-wide"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_7POINT1_WIDE;
      else if (!FFSTRCASECMP(FFL("layout-7point1-wide-back"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_7POINT1_WIDE_BACK;
      else if (!FFSTRCASECMP(FFL("layout-octagonal"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_OCTAGONAL;
      else if (!FFSTRCASECMP(FFL("layout-hexadecagonal"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_HEXADECAGONAL;
      else if (!FFSTRCASECMP(FFL("layout-stereo-downmix"),optarg))
        param.decode.request.channel_layout=AV_CH_LAYOUT_STEREO_DOWNMIX;
      else {
        FFVMESSAGE(FFL("argument to option --matrix not recognized: \"%s\""),
            optarg);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
#endif // ]
    case BG_ARG_STEREO:
#if defined (BG_PARAM_STEREO) // [
      param.stereo=1;
#else // ] [
      param.decode.request.channel_layout=AV_CH_LAYOUT_STEREO;
#endif // ]
      break;
#if defined (BG_PARAM_SEQUENTIAL) // [
    case BG_ARG_SEQUENTIAL:
      param.sequential=1;
      break;
#endif // ]
    /////////////////////////////////////////////////////////////////////////
    case BG_ARG_MOMENTARY_RANGE:
      param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_RANGE;
      break;
    ////////
    case BG_ARG_MOMENTARY_LENGTH:
      param.momentary.ms=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_OVERLAP:
      overlap=FFATOF(optarg);

      if (0.0<=overlap&&overlap<100.0)
        param.momentary.partition=floor(100.0/(100.0-overlap)+0.5);
      else {
        FFFPUTS(FFL("Error: overlap out of range [0..100).\n\n"),stderr);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    case BG_ARG_MOMENTARY_MEAN_GATE:
      param.momentary.mean.gate=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_RANGE_GATE:
      param.momentary.range.gate=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_RANGE_LOWER_BOUND:
      param.momentary.range.lower_bound=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_RANGE_UPPER_BOUND:
      param.momentary.range.upper_bound=FFATOF(optarg);
      break;
    /////////////////////////////////////////////////////////////////////////
    case BG_ARG_SHORTTERM_MEAN:
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MEAN;
      break;
    ////////
    case BG_ARG_SHORTTERM_LENGTH:
      param.shortterm.ms=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_OVERLAP:
      overlap=FFATOF(optarg);

      if (0.0<=overlap&&overlap<100.0)
        param.shortterm.partition=floor(100.0/(100.0-overlap)+0.5);
      else {
        FFFPUTS(FFL("Error: overlap out of range [0..100).\n\n"),stderr);
        bg_usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    case BG_ARG_SHORTTERM_MEAN_GATE:
      param.shortterm.mean.gate=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_RANGE_GATE:
      param.shortterm.range.gate=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_RANGE_LOWER_BOUND:
      param.shortterm.range.lower_bound=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_RANGE_UPPER_BOUND:
      param.shortterm.range.upper_bound=FFATOF(optarg);
      break;
    default:
      if ('c'==optopt)
        FFVMESSAGE(FFL("option -%c requires an argument"),optopt);
      else if (isprint(optopt))
        FFVMESSAGE(FFL("unknown option -%c"),optopt);
      else
        FFVMESSAGE(FFL("unknown option character \\x%x"),optopt);

      bg_usage(argv[0],stderr);
      goto e_arg;
    }
  }

  // in case no class of tags is given we provide both, i.e. album and
  // track tags.
  if (!(param.flags.mode&BG_FLAGS_MODE_TAGS_ALL))
    param.flags.mode|=BG_FLAGS_MODE_TAGS_ALL;

#if defined (HAVE_FF_DYNLOAD) // [
  /////////////////////////////////////////////////////////////////////////////
  // load the FFmpeg libraries from "bs1770gain-tools".
  if (ff_dynload(TOOLS)<0) {
    PBU_DMESSAGE("loading shared libraries");
    goto e_dynload;
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  version=avutil_version();

  if (LIBAVUTIL_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of avutil: %d needed, %d loaded from \"%s\"",
        LIBAVUTIL_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of avutil: %d needed, %d linked",
        LIBAVUTIL_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  ////////
  version=swresample_version();

  if (LIBSWRESAMPLE_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBSWRESAMPLE_VERSION_MAJOR,AV_VERSION_MAJOR(version),
        ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBSWRESAMPLE_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  ////////
  version=avcodec_version();

  if (LIBAVCODEC_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of avcodec: %d needed, %d loaded from \"%s\"",
        LIBAVCODEC_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of avcodec: %d needed, %d linked",
        LIBAVCODEC_VERSION_MAJOR,version);
#endif // ]
    goto e_versions;
  }

  ////////
  version=avformat_version();

  if (LIBAVFORMAT_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of avformat: %d needed, %d loaded from \"%s\"",
        LIBAVFORMAT_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of avformat: %d needed, %d linked",
        LIBAVFORMAT_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  ////////
  version=swscale_version();

  if (LIBSWSCALE_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBSWSCALE_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBSWSCALE_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  ////////
  version=postproc_version();

  if (LIBPOSTPROC_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBPOSTPROC_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBPOSTPROC_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  ////////
  version=avfilter_version();

  if (LIBAVFILTER_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBAVFILTER_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBAVFILTER_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  /////////////////////////////////////////////////////////////////////////////
#if (LIBAVFORMAT_VERSION_MAJOR<58) // [
  av_register_all();
#endif // ]
#if (LIBAVFILTER_VERSION_MAJOR<7) // [
  avfilter_register_all();
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (help_args) {
    if (BG_FLAG_HELP_ARG_NULL&help_args) {
      /////////////////////////////////////////////////////////////////////////
      bg_usage(argv[0],stdout);
    }

    if (BG_FLAG_HELP_ARG_SUFFIX&help_args) {
      /////////////////////////////////////////////////////////////////////////
      fprintf(stdout,"available arguments for option --suffix:\n");
#if (LIBAVFORMAT_VERSION_MAJOR<58) // [
      // https://stackoverflow.com/questions/2940521/where-to-get-full-list-of-libav-formats
      oformat=av_oformat_next(NULL);

      while (oformat) {
        if (oformat->audio_codec&&oformat->extensions)
          fprintf(stdout,"  %s\n",oformat->extensions);

        oformat=av_oformat_next(oformat);
      }
#else // ] [
      opaque=NULL;

      for (;;) {
        oformat=av_muxer_iterate(&opaque);

        if (!oformat)
          break;

        if (oformat->audio_codec&&oformat->extensions)
          fprintf(stdout,"  %s\n",oformat->extensions);
      }
#endif // ]
    }

    if (BG_FLAG_HELP_ARG_CODEC&help_args) {
      /////////////////////////////////////////////////////////////////////////
      fprintf(stdout,"available arguments for option --codec:\n");
#if (LIBAVCODEC_VERSION_MAJOR<58) // [
      // https://stackoverflow.com/questions/2940521/where-to-get-full-list-of-libav-formats
      codec=av_codec_next(NULL);

      while (codec) {
        if (AVMEDIA_TYPE_AUDIO==codec->type)
          fprintf(stdout,"  %s\n",codec->name);

        codec=av_codec_next(codec);
      }
#else // ] [
      opaque=NULL;

      for (;;) {
        codec=av_codec_iterate(&opaque);

        if (!codec)
          break;

        if (AVMEDIA_TYPE_AUDIO==codec->type)
          fprintf(stdout,"  %s\n",codec->name);
      }
#endif // ]
    }

    ///////////////////////////////////////////////////////////////////////////
    code=0;
    goto e_help;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (_WIN32) // [
#if defined (BG_WIN32_CREATE_LOCALE) // [
  param.locale=_create_locale(LC_ALL,"");
#endif // ]
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // e.g. basename.
  param.oem=!lang||!strstr(lang,"UTF-8");
#else // ] [
  // If not already active we set locale to an arbitrary UTF-8.
  locale=setlocale(LC_ALL,NULL);

  if (strstr(locale,"utf")||strstr(locale,"UTF"))
    locale=NULL;
  else
    setlocale(LC_ALL,"en_US.UTF-8");
#endif // ]


  /////////////////////////////////////////////////////////////////////////////
  if (fpath) {
#if defined (_WIN32) // [
    _wremove(fpath);

    if (utf16) {
      param.result.f=_wfopen(fpath,L"wt,ccs=UTF-16LE");
      param.print.vtbl->encoding(&param,16);
    }
    else {
      param.result.f=_wfopen(fpath,L"wt,ccs=UTF-8");
      param.print.vtbl->encoding(&param,8);
    }
#else // ] [
    remove(fpath);
    param.result.f=fopen(fpath,"w");
    param.print.vtbl->encoding(&param,8);
#endif // ]

    if (!param.result.f) {
#if defined (_WIN32) // [
      DVMESSAGEW(L"opening file \"%s\"",fpath);
#else // ] [
      DVMESSAGE("opening file \"%s\"",fpath);
#endif // ]
      goto e_file;
    }
  }
#if defined (BG_PARAM_DUMP) // [
  else if (param.dump)
    param.result.f=stderr;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  av_log_set_level(param.loglevel);

  /////////////////////////////////////////////////////////////////////////////
  if (optind==argc) {
    fputs("Error: nothing to analyze.\n\n",stderr);
    bg_usage(argv[0],stderr);
    goto e_analyzer;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_create(&analyzer)<0) {
    DMESSAGE("creating analyzer");
    goto e_analyzer;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_PARAM_SEQUENTIAL) // [
  if (param.sequential&&(param.output.dirname||param.overwrite)) {
#else // ] [
  if (param.output.dirname||param.overwrite) {
#endif // ]
    if (bg_muxer_create(&muxer)<0) {
      DMESSAGE("creating muxer");
      goto e_muxer;
    }
#if ! defined (BG_PARAM_SEQUENTIAL) // [
  }
#else // ] [
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (!param.suppress.progress) {
#if defined (FF_PRINTER) // [
    fputs("scanning ",stdout);
#else // ] [
    fputs("analyzing ...\n",stdout);
#endif // ]
    fflush(stdout);
  }

  if (bg_root_create(&root,&param)<0) {
    DMESSAGE("creating root");
    goto e_root;
  }

  while (optind<argc) {
#if defined (BG_TREE_CREATE_CHILD_WARNING) // [
    if (bg_root_append(&root,bg_pathnorm(argv[optind]))<0)
      BGVWARNING(FFL("appending \"%s\" to root"),argv[optind]);
#else // ] [
    bg_root_append(&root,bg_pathnorm(argv[optind]));
#endif // ]
    ++optind;
  }

#if defined (FF_PRINTER) // [
  if (!param.suppress.progress) {
    fputs("\nanalyzing ...\n",stdout);
    ff_printer_clear(&param.printer);
    fflush(stdout);
  }
#endif // ]

  if (analyzer.vtbl&&root.vtbl->accept(&root,&analyzer)<0) {
    DMESSAGE("analyzing");
    goto e_analyzer_accept;
  }

  fflush(stdout);

  if (muxer.vtbl) {
    ///////////////////////////////////////////////////////////////////////////
    if (!param.suppress.progress) {
      fputs("remuxing ...\n",stdout);
      fflush(stdout);
    }

    if (root.vtbl->accept(&root,&muxer)<0) {
      DMESSAGE("muxing phase");
      goto e_muxerer_accept;
    }

    fflush(stdout);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (!param.suppress.progress) {
    fputs("done.\n",stdout);
    fflush(stdout);
  }

  code=0;
//cleanup:
e_muxerer_accept:
e_analyzer_accept:
  root.vtbl->destroy(&root);
e_root:
  if (muxer.vtbl)
    muxer.vtbl->destroy(&muxer);
e_muxer:
  if (analyzer.vtbl)
    analyzer.vtbl->destroy(&analyzer);
e_analyzer:
  if (fpath)
    fclose(param.result.f);
e_file:
#if defined (_WIN32) // [
#if defined (BG_WIN32_CREATE_LOCALE) // [
  _free_locale(param.locale);
#endif // ]
#else // ] [
  if (locale)
    setlocale(LC_ALL,locale);
#endif // ]
e_help:
e_versions:
#if defined (HAVE_FF_DYNLOAD) // [
  ff_unload();
e_dynload:
#endif // ]
  bg_param_destroy(&param);
e_arg:
  return code;
}
